module Lexer exposing (..)


import Parser exposing (Parser, (|.), (|=), symbol, succeed, oneOf, variable,
                        sequence, spaces, int, backtrackable)
import Set



-- Tokens:
-- OpenParen    := '('
-- CloseParen   := ')'
-- Quote        := '\''
-- Number       := [ '-' ] ( NonZeroDigit { Digit } | '0' )
-- Identifier   := IdChar { IdChar }
-- Digit        := '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'
-- NonZeroDigit := Digit \ '0'
-- IdChar       := Any \ Digit \ ')' \ '(' \ '\'' \ Whitespace
-- Whitespace   := '\n' | '\r' | ' '



type Token
  = OpenParen
  | CloseParen
  | Quote
  | Number Int
  | Identifier String


lexer : Parser (List Token)
lexer = sequence
  { start = ""
  , separator = ""
  , end = ""
  , spaces = spaces
  , item = lexToken
  , trailing = Parser.Optional
  } 


lexToken : Parser Token
lexToken = oneOf
  [ succeed OpenParen
    |. symbol "("
  , succeed CloseParen
    |. symbol ")" 
  , succeed Quote
    |. symbol "'" 
  , succeed Number
    |= backtrackable myInt
  , succeed Identifier
    |= myIdentifier
  ]


myIdentifier : Parser String
myIdentifier = variable
  { start = isIdentifierChar
  , inner = isIdentifierChar
  , reserved = Set.empty
  }


isIdentifierChar : Char -> Bool
isIdentifierChar c =
  String.contains (String.fromChar c) "0123456789()' \n\r" |> not
  -- Fuck tabs
  -- ~ John Elm, circa 20.19


myInt : Parser Int
myInt =
  oneOf
    [ succeed negate
        |. symbol "-"
        |= int
    , int
    ]
