macro_rules! lisp {
    (($($a:tt)+) $($rest:tt)*) => {
        lisp!(@expand_ln $($a)+);
        $(lisp!($rest);)*
    };

    (($a:literal)) => { $a };

    (@expand_ln + $($t:tt)+) => { lisp!(@expand_op + $($t)+) };
    (@expand_ln - $($t:tt)+) => { lisp!(@expand_op - $($t)+) };
    (@expand_ln * $($t:tt)+) => { lisp!(@expand_op * $($t)+) };
    (@expand_ln = $($t:tt)+) => { lisp!(@expand_op == $($t)+) };
    (@expand_ln > $($t:tt)+) => { lisp!(@expand_op > $($t)+) };
    (@expand_ln < $($t:tt)+) => { lisp!(@expand_op < $($t)+) };

    (@expand_ln print $($a:tt)+) => {
        println!("{}", lisp!(@expand_args $($a)+))
    };
  
    (@expand_ln defun $name:ident ($($args:ident)+) $($t:tt)*) => {
        let $name = |$($args: isize),+| {
            $(;lisp!($t))*
        };
    };

    (@expand_ln defvar $name:ident $($val:tt)+) => {
        let mut $name = lisp!(@expand_arg $($val)+);
    };

    (@expand_ln setvar $name:ident $($val:tt)+) => { $name = lisp!(@expand_arg $($val)+); };

    (@expand_ln if ($($cond:tt)+) $(then)? ($($then:tt)+) $(else)? $(($($else:tt)+))?) => {
        if lisp!(@expand_ln $($cond)+) {
            lisp!(@expand_ln $($then)+)
        } else {
            $(lisp!(@expand_ln $($else)+))?
        }
    };

    (@expand_ln dotimes ($i:ident $($times:tt)+) $($t:tt)+) => {
        for $i in 0..lisp!(@expand_arg $($times)+) {
            lisp!($($t)+);
        }
    };

    (@expand_ln loop $(while)? ($($cond:tt)+) $(do)? $($t:tt)+) => {
        while lisp!(@expand_ln $($cond)+) {
            lisp!($($t)+);
        }
    };

    (@expand_ln return $($t:tt)+) => {
        return lisp!(@expand_arg $($t)+)
    };

    (@expand_ln $fn:ident $($args:tt)*) => {
        $fn($(lisp!(@expand_args $args)),*);
    };

    (@expand_arg $a:literal) => { $a };
    (@expand_arg $a:ident) => { $a };
    (@expand_arg ($($a:tt)+)) => { lisp!(@expand_ln $($a)+) };

    (@expand_args $a:literal $($rest:tt)*) => { $a $(lisp!(@expand_args $rest)),* };
    (@expand_args $a:ident $($rest:tt)*) => { $a $(lisp!(@expand_args $rest)),* };
    (@expand_args ($($n:tt)+) $($rest:tt)*) => { lisp!(@expand_ln $($n)+) $(lisp!(@expand_args $rest)),* };

    (@expand_op $op:tt $n:ident $($rest:tt)+) => { $n $op lisp!(@expand_op $op $($rest)+) };
    (@expand_op $op:tt $n:literal $($rest:tt)+) => { $n $op lisp!(@expand_op $op $($rest)+) };
    (@expand_op $op:tt ($($n:tt)+) $($rest:tt)+) => { lisp!(@expand_ln $($n)+) $op lisp!(@expand_op $op $($rest)+) };
    (@expand_op $op:tt $n:ident) => { $n };
    (@expand_op $op:tt $n:literal) => { $n };
    (@expand_op $op:tt ($($n:tt)+)) => { lisp!(@expand_ln $($n)+) };
}

pub(crate) use lisp;